             .title        "CNROM Reader Plugin"

;06/10/00 
;Written by KH
;Version 1.0

;07/22/06
;BootGod
;added open bus check for first bank as well

;04/28/07
;BootGod
;added user var for diode setting

;07/30/07
;BootGod
;Complete rewrite, 8K CHR plugin merged into this one, CHR enable user val not needed anymore,
;fixes under-dumping of games like Mighty Bomb Jack (USA)

;09/14/09
;BootGod
;Runs write bank routine before checking mirroring to ensure diode value
;is written in first.

;10/25/09
;BootGod
;Added a special block type to return aux data, in this case, it returns 1 byte containing
;the diode code used and more importantly, the code to enable an 8K CHR ROM.
             
             ;vectors for standard system calls

send_byte:   .equ 0200h
baton:       .equ 0203h
chk_vram:    .equ 0206h
chk_wram:    .equ 0209h
wr_ppu:      .equ 020ch
read_byte:   .equ 020fh
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

crcs:        .equ 00d0h

addr1:       .equ 00e0h
addr1_lo:    .equ 00e0h
addr1_hi:    .equ 00e1h
addr2:       .equ 00e2h
addr2_lo:    .equ 00e2h
addr2_hi:    .equ 00e3h

numbanks:    .equ 00e4h
curbank:     .equ 00e5h
reg:         .equ 00e6h
goodbanks:   .equ 00e7h
banksleft:   .equ 00e8h
is8kchr:     .equ 00e9h

x_temp:      .equ 00eah
y_temp:      .equ 00ebh

uv1enabled:  .equ 07ceh
diodeval:    .equ 07cfh

;plugin header defines
p_ver: = 6	;plugin version
p_read: = 1	;plugin is for downloading (dumping)
p_write: = 2	;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3	;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4	;plugin uses user variable #1
p_var2: = 8	;plugin uses user variable #2
p_var3: = 16	;plugin uses user variable #3
p_var4: = 32	;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "Nintendo CNROM carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "BootGod"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 16 >> 4,32 >> 4
             .db 8 >> 3,32 >> 3
             .db 0,0

             ;plugin version and flags
             .db p_ver
             .db p_read | p_var1

;plugin code start here

             lda #0
             sta is8kchr
             jsr wr_bank	;bank write is to make sure diode value is written before checking mirroring
             
;check mirroring (x = mirroring)
             ldx #00h		;Horizontal
             lda #020h
             jsr wr_ppu
             lda #055h
             sta 2007h
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #055h
             bne got_mir
             lda 2007h
             cmp #0aah
             bne got_mir
             lda #020h
             jsr wr_ppu
             lda #0aah
             sta 2007h
             lda #024h
             jsr wr_ppu
             lda 2007h
             lda 2007h
             cmp #0aah
             bne got_mir
             inx		;Vertical
got_mir:     txa
             jsr send_byte	;send mirroring
             
;determine PRG ROM size (16K or 32K)
             lda #000h
             sta addr1_lo
             sta addr2_lo
             lda #080h
             sta addr1_hi
             lda #0c0h
             sta addr2_hi	;set addresses to $8000 and $C000
             
             ldx #040h
             ldy #0		;set up to check 16K

chk_prg:     lda (addr1),y
             cmp (addr2),y
             bne is32k
             iny
             bne chk_prg
             inc addr1_hi
             inc addr2_hi
             dex
             bne chk_prg

is16k:       lda #040h		;40h * 256bytes = 16K
             tax
             bne send_prg

is32k:       lda #080h		;80h * 256bytes = 32K
             tax

send_prg:    jsr send_byte
             lda #000h
             jsr send_byte
             lda #001h
             jsr send_byte	;send PRG header

             ldy #0
             sty addr1_lo
             lda #080h
             sta addr1_hi	;set address to $8000, X = num blocks to send

dump_prg:    lda (addr1),y
             jsr send_byte
             iny
             bne dump_prg
             inc addr1_hi
             dex
             bne dump_prg
             
;determine CHR ROM size (8K or 16K or 32K)             
do_chr:      lda #2
             sta goodbanks	;clear flag byte (sets which banks should be dumped later)
             lda #0
             sta numbanks	;set num banks to 0

             lda #3
             sta curbank	;start with bank 4 and work down

             lda #4
             sta banksleft

chk_chr:     lda curbank
             jsr wr_bank	;set bank
             jsr crc_bank	;get crc
             beq is_ob		;if open bus, jump below

             ora goodbanks
             sta goodbanks	;flag as good bank

             lda curbank
             asl a
             asl a
             tay		;init index for bank# CRC
             ldx #3		;init index for reading last crc

save_crc:    lda crc0,x		;save the crc
             sta crcs,y
             iny
             dex
             bne save_crc
             
             inc numbanks	;inc num good banks
             
is_ob:       asl goodbanks	;shift flags over 1
             dec curbank
             dec banksleft
             bne chk_chr	;continue to next bank

             lda numbanks	;now that we've checked all 4 banks, if all 4 banks are good,
             cmp #4             ;we need to make sure it's not actually 16K mirrored
             bne chr_ready

             ldx #07h
             ldy #0fh

check_crcs:  lda crcs,x
             cmp crcs,y
             bne chr_ready
             dey
             dex
             bne check_crcs

             lda #6		;banks 0-1 matched 2-3, so fix vars (6 = 110b which will be 11b after corrective shift)
             sta goodbanks
             lda #2
             sta numbanks

chr_ready:   lda numbanks	;convert numbanks to num pages
             cmp #1		;check if is 8K CHR
             bne chr_shift
             sta is8kchr	;it is, so set flag to send special block later
chr_shift:   asl a
             asl a
             asl a
             asl a
             asl a
             jsr send_byte
             lda #0
             jsr send_byte
             lda #02h
             jsr send_byte 	;send CHR header
             
             lsr goodbanks
             lda #0
             sta curbank

next_chr:    lda curbank
             jsr wr_bank	;set bank
             inc curbank

             lsr goodbanks	;check if good bank
             bcc next_chr

             lda #0
             jsr wr_ppu		;set to PPU $0000
             lda 2007h

             ldx #32		;setup for 8K dump
             ldy #0

send_chr:    lda 2007h
             jsr send_byte
             iny
             bne send_chr
             dex
             bne send_chr
             dec numbanks
             bne next_chr

send_spec:   lda is8kchr	;check if was 8K CHR dump
             beq done		;skip if it wasn't
             lda uv1enabled	;also check if user var was entered
             cmp #0ffh
             beq done		;skip if it wasn't so old clients dont choke on unknown block

             lda #1		;size for special block interpreted as bytes instead of
             jsr send_byte	;blocks. In this case we are only sending 1 byte
             lda #0
             jsr send_byte
             lda #10		;block type is defined as 10
             jsr send_byte

             lda reg		;send register value used to dump
             jsr send_byte
             
done:        lda #000h		;send end block
             jsr send_byte
             lda #000h
             jsr send_byte
             lda #000h
             jsr send_byte

             rts		;return to BIOS
             

;sub for setting CHR bank (compensates for bus conflicts)
wr_bank:     stx x_temp
             sty y_temp

             ldx uv1enabled	;check if user has enter value for D4/D5
             cpx #0ffh
             beq no_diodes
             ora diodeval	;if so, OR it with bank num

no_diodes:   sta reg		;store value to be written to reg

             lda #0
             sta addr1_lo
             lda #080h
             sta addr1_hi	;set address to $8000

             ldx #080h		;setup to search all 32K of PRG
             ldy #0

wrb_loop:    lda (addr1),y
             and #033h		;mask it to just the bits we need to match
             cmp reg		;check if current byte matches what we want to write
             beq found_reg
             iny
             bne wrb_loop
             inc addr1_hi
             bne wrb_loop

no_reg:      lda reg		;didn't find place to write, try it anyways
             sta 08000h
             jmp wrb_return

found_reg:   lda (addr1),y	;found value equal to reg, write it.
             sta (addr1),y

wrb_return:  ldx x_temp
             ldy y_temp

             rts

;sub for doing CRC on 8K CHR at $0000, returns 0 if was open bus, 1 if not
crc_bank:    lda #0
             jsr wr_ppu		;set to PPU $0000

             jsr init_crc
             lda 2007h

             ldx #020h		;checking 8K
             ldy #0

crc_loop:    lda 2007h
             stx x_temp
             jsr do_crc
             ldx x_temp
             iny
             bne crc_loop
             dex
             bne crc_loop

             jsr finish_crc

             ldx #3		;check if open bus type 1
ob1_check:   lda crc0,x
             cmp openbus1,x
             bne _ob2check
             dex
             bne ob1_check

is_ob1:      lda #0
             rts

_ob2check:   ldx #3		;check if open bus type 2
ob2_check:   lda crc0,x
             cmp openbus2,x
             bne crc_pass
             dex
             bne ob2_check

is_ob2:      lda #0
             rts

crc_pass:    lda #1
             rts

openbus1:    .db 007h,053h,067h,0b6h
openbus2:    .db 01ah,0c9h,05ch,078h

             .fill 07c0h-*,0ffh   ;fill rest to get 1K of data

             .db "Diode Code"
             .fill 07ceh-*,0
             .fill 07d0h-*,0ffh

             .fill 0800h-*,0ffh
             .end
